from dataclasses import dataclass
from datetime import datetime
from typing import Iterator, Optional
from uuid import uuid4

from injector import inject
from pyckson import parse, serialize, rename
from pymongo.database import Database

from antibot.tools import today


@rename(id='_id')
@dataclass
class SlackMessage:
    id: str
    type: str
    date: datetime
    timestamp: str
    channel_id: Optional[str]

    @staticmethod
    def create_today(type: str, timestamp: str, channel_id: Optional[str] = None) -> 'SlackMessage':
        return SlackMessage(str(uuid4()), type, today(), timestamp, channel_id)


class MessagesRepository:
    @inject
    def __init__(self, db: Database):
        self.collection = db['messages']

    def find_all(self, type: str, date: Optional[datetime] = None) -> Iterator[SlackMessage]:
        query = {'type': type}
        if date is not None:
            query['date'] = date
        for item in self.collection.find(query):
            yield parse(SlackMessage, item)

    def find_one(self, type: str, date: Optional[datetime] = None) -> Optional[SlackMessage]:
        query = {'type': type}
        if date is not None:
            query['date'] = date
        doc = self.collection.find_one(query)
        if doc is not None:
            return parse(SlackMessage, doc)

    def update_timestamp(self, id: str, timestamp: str):
        self.collection.update({'_id': id}, {'$set': {'timestamp': timestamp}})

    def delete(self, id: str):
        self.collection.remove({'_id': id})

    def create(self, message: SlackMessage):
        self.collection.insert_one(serialize(message))



import os
from typing import List, Type

from injector import Module, Binder, singleton, inject
from pymongo.database import Database
from pymongo.mongo_client import MongoClient
from slack import WebClient

from antibot.internal.configuration import Configuration
from antibot.internal.plugins import PluginsCollection
from antibot.plugin import AntibotPlugin


def configuration_provider() -> Configuration:
    return Configuration(bot_user_oauth_token=os.environ['SLACK_BOT_USER_TOKEN'],
                         signing_secret=os.environ['SIGNING_SECRET'],
                         ws_api_key=os.environ['WS_API_KEY'],
                         vhost=os.environ.get('VHOST', 'http://localhost:5001'),
                         prod=not os.environ.get('DEV', False))


@inject
def slack_client_provider(configuration: Configuration) -> WebClient:
    return WebClient(configuration.bot_user_oauth_token)


class AntibotModule(Module):
    def __init__(self, plugins: List[Type[AntibotPlugin]], submodules=List[Type[Module]]):
        self.plugins = plugins
        self.submodules = submodules

    def configure(self, binder: Binder):

        for plugin in self.plugins:
            binder.bind(plugin, scope=singleton)

        binder.bind(Configuration, to=configuration_provider)
        binder.bind(WebClient, to=slack_client_provider)
        binder.bind(PluginsCollection, to=PluginsCollection(self.plugins))

        binder.bind(Database, to=MongoClient(os.environ['MONGO_URI'])['antibot'])

        for module in self.submodules:
            binder.install(module())



from inspect import signature
from threading import Thread
from typing import Type, Callable

from injector import inject, Injector
from pyckson import loads

from antibot.internal.backend.constants import ASYNC_REPLY
from antibot.internal.backend.debugger import Debugger


class EndpointRunner:
    @inject
    def __init__(self, injector: Injector, debugger: Debugger):
        self.injector = injector
        self.debugger = debugger

    def run(self, plugin: Type, method: Callable, **kwargs):
        instance = self.injector.get(plugin)
        method_args = {}

        for name, param in signature(method).parameters.items():
            if name in kwargs:
                if param.annotation != str and type(kwargs[name]) is str:
                    kwargs[name] = loads(param.annotation, kwargs[name])
                method_args[name] = kwargs[name]

        if getattr(method, ASYNC_REPLY, False):
            def runner(r_method, r_instance, r_args):
                with self.debugger.wrap(method_args):
                    r_method(r_instance, **r_args)

            t = Thread(target=runner, args=[method, instance, method_args])
            t.start()
            return
        return method(instance, **method_args)



import re
from dataclasses import dataclass
from typing import Type, Callable, Iterable

from injector import singleton, inject
from pyckson import parse

from antibot.internal.backend.constants import BLOCK_ACTION_OPTIONS
from antibot.internal.backend.descriptor import find_method_by_attribute
from antibot.internal.backend.endpoint_runner import EndpointRunner
from antibot.internal.slack.channel import Channel
from antibot.plugin import AntibotPlugin
from antibot.repository.users import UsersRepository
from antibot.slack.api import SlackApi
from antibot.slack.callback import BlockPayload
from antibot.slack.message import Message


@dataclass
class BlockActionDescriptor:
    plugin: Type[AntibotPlugin]
    method: Callable
    block_id: str
    action_id: str


@singleton
class BlockActionRunner:
    @inject
    def __init__(self, endpoints: EndpointRunner, users: UsersRepository, api: SlackApi):
        self.endpoints = endpoints
        self.users = users
        self.api = api
        self.block_actions = []

    def install_plugin(self, plugin: Type[AntibotPlugin]):
        for method, options in find_method_by_attribute(plugin, BLOCK_ACTION_OPTIONS):
            self.block_actions.append(BlockActionDescriptor(plugin, method, options.block_id, options.action_id))

    def find_block_action(self, block_id: str, action_id: str) -> Iterable[BlockActionDescriptor]:
        for block_action in self.block_actions:
            if block_action.block_id is not None:
                if block_action.block_id == block_id or re.match(block_action.block_id, block_id):
                    yield block_action
                    continue
            if block_action.action_id is not None:
                if block_action.action_id == action_id or re.match(block_action.action_id, action_id):
                    yield block_action
                    continue

    def run_callback(self, payload: dict):
        message = parse(BlockPayload, payload)
        for action in message.actions:
            for block_action in self.find_block_action(action.block_id, action.action_id):
                user = self.users.get_user(message.user.id)
                channel = Channel(message.channel.id, message.channel.name) if message.channel else None
                values = message.view.state.values if message.view and message.view.state else None
                reply = self.endpoints.run(block_action.plugin, block_action.method,
                                           user=user, channel=channel,
                                           action=action, trigger_id=message.trigger_id,
                                           timestamp=message.container.message_ts,
                                           response_url=message.response_url,
                                           view_id=message.view.id if message.view else None,
                                           private_metadata=message.view.private_metadata if message.view else None,
                                           values=values)

                if isinstance(reply, Message):
                    self.api.respond(message.response_url, reply)



from dataclasses import dataclass
from typing import Type, Callable, Iterable

from injector import inject, singleton
from pyckson import parse

from antibot.internal.backend.constants import VIEW_CLOSED_ID
from antibot.internal.backend.descriptor import find_method_by_attribute
from antibot.internal.backend.endpoint_runner import EndpointRunner
from antibot.plugin import AntibotPlugin
from antibot.repository.users import UsersRepository
from antibot.slack.api import SlackApi
from antibot.slack.callback import ViewClosedPayload


@dataclass
class ViewClosedDescriptor:
    plugin: Type[AntibotPlugin]
    method: Callable
    callback_id: str


@singleton
class ViewClosedRunner:
    @inject
    def __init__(self, endpoints: EndpointRunner, users: UsersRepository, api: SlackApi):
        self.endpoints = endpoints
        self.users = users
        self.api = api
        self.descriptors = []

    def install_plugin(self, plugin: Type[AntibotPlugin]):
        for method, callback_id in find_method_by_attribute(plugin, VIEW_CLOSED_ID):
            self.descriptors.append(ViewClosedDescriptor(plugin, method, callback_id))

    def find_callback(self, callback_id) -> Iterable[ViewClosedDescriptor]:
        for descriptor in self.descriptors:
            if descriptor.callback_id == callback_id:
                yield descriptor

    def run(self, payload: dict):
        message = parse(ViewClosedPayload, payload)
        for descriptor in self.find_callback(message.view.callback_id):
            user = self.users.get_user(message.user.id)
            self.endpoints.run(descriptor.plugin, descriptor.method,
                               user=user, callback_id=message.view.callback_id,
                               private_metadata=message.view.private_metadata)



CMD_ATTR = '__cmd'
CALLBACK_ID_REGEX = '__callback'
BLOCK_ACTION_OPTIONS = '__block_action'
DIALOG_SUBMIT_ID = '__dialog_submit'
DIALOG_CANCEL_ID = '__dialog_cancel'
JOB_ATTR_DAILY = '__schedule'
WS_ATTR = '__ws'
WS_JSON_VALUES = '__ws_json'
VIEW_CLOSED_ID = '__view_closed'
VIEW_SUBMIT_ID = '__view_submit'
ASYNC_REPLY = '__async_reply'




from typing import Optional

from antibot.internal.backend.constants import CMD_ATTR, JOB_ATTR_DAILY, WS_ATTR, WS_JSON_VALUES, BLOCK_ACTION_OPTIONS, \
    VIEW_CLOSED_ID, VIEW_SUBMIT_ID, ASYNC_REPLY
from antibot.internal.backend.descriptor import CommandDescriptor, WsDescriptor, BlockActionOptions


def command(route):
    def decorator(f):
        setattr(f, CMD_ATTR, CommandDescriptor(route, f))
        return f

    return decorator


def block_action(block_id: Optional[str] = None, action_id: Optional[str] = None):
    def decorator(f):
        setattr(f, BLOCK_ACTION_OPTIONS, BlockActionOptions(block_id, action_id))
        return f

    return decorator


def view_closed(callback_id: str):
    def decorator(f):
        setattr(f, VIEW_CLOSED_ID, callback_id)
        return f

    return decorator


def view_submit(callback_id: str):
    def decorator(f):
        setattr(f, VIEW_SUBMIT_ID, callback_id)
        return f

    return decorator


def ws(route, method='POST'):
    def decorator(f):
        setattr(f, WS_ATTR, WsDescriptor(route, method, f))
        return f

    return decorator


def jsonobject():
    def decorator(f):
        setattr(f, WS_JSON_VALUES, True)
        return f

    return decorator


def daily(hour='00:00'):
    def decorator(f):
        setattr(f, JOB_ATTR_DAILY, hour)
        return f

    return decorator


def async_reply(f):
    setattr(f, ASYNC_REPLY, True)
    return f



from datetime import datetime
from typing import Iterable

import arrow

from antibot.slack.message import Message, Block


def today() -> datetime:
    return arrow.utcnow().floor('day').datetime


def yesterday() -> datetime:
    return arrow.utcnow().floor('day').shift(days=-1).datetime


def message(blocks: Iterable[Block], **kwargs) -> Message:
    return Message(blocks=list(blocks), **kwargs)



import logging
from functools import partial
from typing import Type

from bottle import route
from injector import inject

from antibot.internal.backend.actions.action_runner import ActionRunner
from antibot.internal.backend.actions.block_action_runner import BlockActionRunner
from antibot.internal.backend.actions.view_closed_runner import ViewClosedRunner
from antibot.internal.backend.actions.view_submit_runner import ViewSubmitRunner
from antibot.internal.backend.command_runner import CommandRunner
from antibot.internal.backend.descriptor import find_commands, find_ws
from antibot.internal.backend.ws_runner import WsRunner
from antibot.internal.configuration import Configuration
from antibot.plugin import AntibotPlugin


class PluginInstaller:
    @inject
    def __init__(self, cmd_runner: CommandRunner, configuration: Configuration, block_action_runner: BlockActionRunner,
                 ws_runner: WsRunner, action_runner: ActionRunner, view_closed_runner: ViewClosedRunner,
                 view_submit_runner: ViewSubmitRunner):
        self.cmd_runner = cmd_runner
        self.configuration = configuration
        self.block_action_runner = block_action_runner
        self.ws_runner = ws_runner
        self.action_runner = action_runner
        self.view_closed_runner = view_closed_runner
        self.view_submit_runner = view_submit_runner

    def install_plugin(self, plugin: Type[AntibotPlugin]):
        for command in find_commands(plugin):
            logging.getLogger(__name__).info('Installing command {}{}'.format(self.configuration.vhost, command.route))
            route(command.route, method='POST')(
                partial(self.cmd_runner.run_command, method=command.method, plugin=plugin))

        self.block_action_runner.install_plugin(plugin)
        self.view_closed_runner.install_plugin(plugin)
        self.view_submit_runner.install_plugin(plugin)

        for ws in find_ws(plugin):
            route(ws.route, method=ws.http_method)(partial(self.ws_runner.run_ws, method=ws.method, plugin=plugin))

        route('/action', method='POST')(lambda: self.action_runner.run())



from dataclasses import dataclass
from typing import Optional

from pyckson import no_camel_case


@no_camel_case
@dataclass
class Profile:
    display_name: str
    real_name: str
    email: Optional[str]


@no_camel_case
@dataclass
class Member:
    id: str
    profile: Profile



import re

from setuptools import setup, find_packages

with open('src/antibot/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)
    assert version is not None

setup(name='antibot',
      version=version,
      author='Jean Giard',
      license='LGPL',
      classifier=[
          'Programming Language :: Python :: 3'
      ],
      entry_points={
          'console_scripts': ['antibot=antibot.main:run'],
          'antibot': [
              'antibot=antibot.provided:BasePlugin',
              'dismiss=antibot.provided:DismissActionPlugin',
              'debug=antibot.provided:DebuggerPlugin',
          ]
      },
      packages=find_packages(where='src'),
      package_dir={'': 'src'},
      install_requires=[
          'pymongo',
          'requests',
          'injector',
          'pyckson',
          'schedule',
          'slackclient',
          'arrow',
          'bottle'
      ],
      )



import hashlib
import hmac
import logging
import time

from bottle import BaseRequest, abort
from injector import inject

from antibot.internal.configuration import Configuration


class RequestChecker:
    @inject
    def __init__(self, configuration: Configuration):
        self.configuration = configuration

    def check_request(self, request: BaseRequest):
        try:
            timestamp = request.headers['X-Slack-Request-Timestamp']
            if abs(time.time() - float(timestamp)) > 60 * 5:
                abort(401, 'Invalid timestamp')
            body = request.body.read()
            data = 'v0:{}:'.format(timestamp).encode() + body
            request_hash = 'v0=' + hmac.new(str.encode(self.configuration.signing_secret),
                                            data, hashlib.sha256).hexdigest()
            signature = request.headers['X-Slack-Signature']
            if not hmac.compare_digest(request_hash, signature):
                abort(401)
        except KeyError:
            logging.error('Could not verify request')
            logging.error(str(dict(request.headers)))
            logging.error(str(dict(request.environ)))
            logging.error(str(dict(request.params)))
            abort(401, 'no X-Slack headers found')
